/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "common.h"
#include "options.h"
#include "conf.h"
#include "log.h"
#include "crypt.h"
#include "thread.h"
#include "parsecfg.h"


void conf_init() {
	extern sig_atomic_t __conf_initialized;
	extern conf_t conf;
	
	object_init(&conf,NULL,NULL);

	/* store static string pointers so we don't free() then */
	conf.__random		= DEFAULT_RANDOMNESS_SOURCE;
	conf.__user		= NULL;
	conf.__group		= NULL;
	conf.__prime		= DEFAULT_PRIME;

	/* defaults */
#define CONFIG SYSCONFDIR ## "/" ## PACKAGE ## "rc"
	conf.config		= (options.config == NULL)? CONFIG : options.config;
#undef CONFIG
	conf.daemon		= true;
	conf.port		= DEFAULT_PORT;
	conf.listen		= NULL;
	conf.verbosity		= DEFAULT_VERBOSITY;
	conf.random		= conf.__random;
	conf.user		= conf.__user;
	conf.group		= conf.__group;
	conf.prime		= conf.__prime;
	conf.enable_error	= (ANET_LOG_ERROR < DEFAULT_VERBOSITY);
	conf.enable_warn	= (ANET_LOG_WARN  < DEFAULT_VERBOSITY);
	conf.enable_info	= (ANET_LOG_INFO  < DEFAULT_VERBOSITY);
	conf.enable_debug	= (ANET_LOG_DEBUG < DEFAULT_VERBOSITY);
	conf.enable_dump	= (ANET_LOG_DUMP  < DEFAULT_VERBOSITY);
	conf.enable_track_source	= false;
	conf.enable_track_func		= false;
	conf.enable_track_thread	= false;

	__conf_initialized = true;
	
	return;
}

void conf_destroy() {
	extern conf_t conf;

	object_destroy(&conf);

}


void conf_parse() {
	extern options_t options;
	extern conf_t conf;
	unsigned int port = 0;
	bool daemon = false;
	char *listen;
	char *random;

	object_lock(&conf);


	/* make sure we don't free static memory on multiple parses */
	if (conf.random != NULL	&& conf.random != conf.__random && conf.random != options.random)
		free(conf.random);
	if (conf.user != NULL	&& conf.user != conf.__user && conf.user != options.user)
		free(conf.user);
	if (conf.group != NULL	&& conf.group != conf.__group && conf.group != options.group)
		free(conf.group);

	
	INFO("reading configuration from %s",conf.config);

	if (options.foreground) {
		conf.daemon = false;
	} else if (options.daemon) {
		conf.daemon = true;
	} else {
		if (fetchVarFromCfgFile(conf.config,"daemon",&daemon,CFG_BOOL,CFG_SIMPLE,0,NULL) == 0) {
			conf.daemon = daemon;
		} else {
			WARN("defaulting to %s.",(conf.daemon)? "run as a daemon": "run in the foreground");
		}
	}


	if (options.port != 0) {
		conf.port = options.port;
	} else {
		if (fetchVarFromCfgFile(conf.config,"port",&port,CFG_UINT,CFG_SIMPLE,0,NULL) == 0) {
			if (0 < port && port < 65536) {
				conf.port = (unsigned short) port;
			} else {
				ERROR("illegal port found in configuration file: %i",port);
				WARN("using default port %i",conf.port);
			}
		} else {
			WARN("using default port %i",conf.port);
		}
	}


	if (options.listen != NULL) {
		conf.listen = options.listen;
	} else {
		if (fetchVarFromCfgFile(conf.config,"listen",&listen,CFG_STRING,CFG_SIMPLE,0,NULL) == 0) {
			conf.listen = strdup(listen);
		}
	}


	if (options.random != NULL) {
		conf.random = options.random;
	} else {
		if (fetchVarFromCfgFile(conf.config,"random",&random,CFG_STRING,CFG_SIMPLE,0,NULL) == 0) {
			conf.random = strdup(random);
		} else {
			WARN("defaulting to random device %s.",conf.random);
		}
	}


	conf.verbosity = MIN(MAX_VERBOSITY,(options.verbosity == 0)? DEFAULT_VERBOSITY : options.verbosity);
	
	switch (conf.verbosity) {
		case 5:
			conf.enable_dump	= true;
		case 4:
			conf.enable_debug	= true;
		case 3:
			conf.enable_info	= true;
		case 2:
			conf.enable_warn	= true;
		case 1:
		default:
			conf.enable_error	= true;
	}


	conf.enable_track_source	= options.track_source;
	conf.enable_track_func		= options.track_func;
	conf.enable_track_thread	= options.track_thread;


	if (options.user != NULL) {
		conf.user = (char *)strdup(options.user);
	}
	
	if (options.group != NULL) {
		conf.group = (char *)strdup(options.group);
	}

	object_unlock(&conf);
		
	return;
}


void conf_reload() {
	conf_parse();
}
