/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>

#ifdef __USE_MCRYPT
#ifndef _MCRYPT_H
#define _MCRYPT_H
#include <mcrypt.h>
#endif
#endif

#ifdef HAVE_GMP_H
#include <gmp.h>
#else
#error No gmp.h!
#endif /* HAVE_GMP_H */

#include "common.h"
#include "conf.h"
#include "log.h"
#include "object.h"
#include "bignum.h"
#include "crypt.h"
#include "sha1.h"
#include "tiger.h"
#include "cbc.h"
#include "job.h"

void rand_init() {
	struct stat *st;
	char *path;
	int *fd;
	unsigned char buf[256];
	int in;


	object_init(&__rand_source,NULL,NULL);
	object_lock(&__rand_source);

	st = &__rand_source.st;
	fd = &__rand_source.fd;

	object_lock(&conf);
		path = __rand_source.path = strdup(conf.random);
	object_unlock(&conf);

	if (stat(path,st) != 0) {
		ERROR("could not stat randomness source %s: %s",path,strerror(errno));
		goto init_prng;
	}

	if (S_ISCHR(st->st_mode)) {
		INFO("assuming %s is a non-blocking character device",path);
		if ((*fd = open(path,O_RDONLY)) == -1) {
			ERROR("could not open randomness source %s: %s",path,strerror(errno));
			goto init_prng;
		} else {
			INFO("opened %s as randomness source",path);
		}
	} else if (S_ISSOCK(st->st_mode)) {
		INFO("assuming %s is an EGD/PRNGD, non-blocking socket",path);
		if ((*fd = sopen(path)) == -1) {
			ERROR("could not open randomness source %s: %s",path,strerror(errno));
			goto init_prng;
		} else {
			INFO("opened %s as randomness source",path);
		}
	} else {
		ERROR("%s is an unknown randomness source",path);
		goto init_prng;
	}

	object_unlock(&__rand_source);
	return;

	init_prng:
	ASSERT(0,"error initializing randomness routines");

	object_unlock(&__rand_source);

	return; /* void */
}


void rand_get(void *buf, size_t bytes) {
	int in;
	size_t interleave;
	unsigned char req[2]; /* PRNGD, EGD request */
	char *_buf = (char *)buf;

	req[0] = 1;
	req[1] = MIN(255,bytes);

	while (bytes > 0) {
		object_lock(&__rand_source);

		if (S_ISSOCK(__rand_source.st.st_mode)) {
			if (write(__rand_source.fd,req,sizeof(req)) != sizeof(req)) {
				ERROR("error writing to socket %s: %s",__rand_source.path,strerror(errno));
				break;
			} else {
				interleave = 255;
			}
		} else {
			interleave = PIPE_BUF;
		}

		while(MIN(bytes,interleave) > 0) {

			if ((in = read(__rand_source.fd,_buf,MIN(bytes,interleave))) <= 0) {
				WARN("error reading from random source %s: %s",__rand_source.path,strerror(errno));
				goto internal_prng;
			}
			bytes		-= in;
			_buf		+= in;
			interleave	-= in;
		}

		object_unlock(&__rand_source);
	}

	

	if (bytes != 0) {
		internal_prng:
		WARN("unable to get external randomness, using internal");
	}

	
	return; /* void */
}


cipher_ctx *cipher_create(int type,void *key,void *iv) {
	cipher_ctx *cipher;
	
	if ((cipher = _malloc(sizeof(cipher_ctx))) == NULL) {
		ERROR("could not malloc space for a new cipher");
		return NULL;
	}
	bzero(cipher,sizeof(cipher_ctx));

	cipher_init(cipher,type,key,iv);

	__objects_created++;
	
	return cipher;
}


void cipher_free(cipher_ctx *cipher) {
	cipher_destroy(cipher);
	
	_free(cipher);

	__objects_freed++;
	
	return;
}


#ifdef __USE_MCRYPT
void __cipher_init() {
	__cipher_init_mutex = object_create();
}
#endif


void cipher_init(cipher_ctx *ctx,int type,void *key,void *iv) {
	static object_t 	o;

	#ifdef __USE_MCRYPT
		pthread_once(&__cipher_init_once,__cipher_init);
		object_lock(__cipher_init_mutex);
	#endif

	switch ((ctx->type = type)) {
		case RIJNDAEL_128:
			#ifdef __USE_MCRYPT

			ASSERT((ctx->__ctx.mcrypt_td = mcrypt_module_open(MCRYPT_RIJNDAEL_128,NULL,MCRYPT_CBC,NULL)) != MCRYPT_FAILED,
				"failed to open mcrypt cipher %s",MCRYPT_RIJNDAEL_128);

			#else

			bzero(ctx->__xor_block,16);
			memcpy(ctx->__xor_block,iv,16);
			aes_set_key((ctx->__ctx_ptr = &ctx->__ctx.aes),16,key);
			ctx->__encrypt =
				(void(*)(void *,unsigned int,unsigned char *,const unsigned char *))aes_encrypt;
			ctx->__decrypt =
				(void(*)(void *,unsigned int,unsigned char *,const unsigned char *))aes_decrypt;

			#endif
			break;
		case TWOFISH_128:
			#ifdef __USE_MCRYPT

			ASSERT((ctx->__ctx.mcrypt_td = mcrypt_module_open(MCRYPT_TWOFISH,NULL,MCRYPT_CBC,NULL)) != MCRYPT_FAILED,
				"failed to open mcrypt cipher %s",MCRYPT_TWOFISH);

			#else

			bzero(ctx->__xor_block,16);
			memcpy(ctx->__xor_block,iv,16);
			twofish_set_key((ctx->__ctx_ptr = &ctx->__ctx.twofish),16,key);
			ctx->__encrypt =
				(void(*)(void *,unsigned int,unsigned char *,const unsigned char *))twofish_encrypt;
			ctx->__decrypt =
				(void(*)(void *,unsigned int,unsigned char *,const unsigned char *))twofish_decrypt;

			#endif
			break;
		case SERPENT_128:
			#ifdef __USE_MCRYPT

			ASSERT((ctx->__ctx.mcrypt_td = mcrypt_module_open(MCRYPT_SERPENT,NULL,MCRYPT_CBC,NULL)) != MCRYPT_FAILED,
				"failed to open mcrypt cipher %s",MCRYPT_SERPENT);

			#else

			bzero(ctx->__xor_block,16);
			memcpy(ctx->__xor_block,iv,16);
			serpent_set_key((ctx->__ctx_ptr = &ctx->__ctx.serpent),16,key);
			ctx->__encrypt =
				(void(*)(void *,unsigned int,unsigned char *,const unsigned char *))serpent_encrypt;
			ctx->__decrypt =
				(void(*)(void *,unsigned int,unsigned char *,const unsigned char *))serpent_decrypt;

			#endif
			break;
		default:
			ABORT("bad cipher requested: %i",type);
	}

	#ifdef __USE_MCRYPT
		ASSERT(mcrypt_generic_init(ctx->__ctx.mcrypt_td,key,16,iv) == 0,
			"failed to initialize cipher");
		object_unlock(__cipher_init_mutex);
	#endif

	return;
}


void cipher_destroy(cipher_ctx *ctx) {
	#ifdef __USE_MCRYPT
		mcrypt_generic_deinit(ctx->__ctx.mcrypt_td);
	#endif
	return;
}


void cipher_encrypt(cipher_ctx *ctx,void *plaintext,int len) {
	#ifdef __USE_MCRYPT

	mcrypt_generic(ctx->__ctx.mcrypt_td,plaintext,len);

	#else

	cbc_encrypt(ctx->__ctx_ptr,ctx->__encrypt,16,ctx->__xor_block,len,plaintext,plaintext);

	#endif

	return;
}


void cipher_decrypt(cipher_ctx *ctx,void *ciphertext,int len) {
	#ifdef __USE_MCRYPT

	mdecrypt_generic(ctx->__ctx.mcrypt_td,ciphertext,len);

	#else

	cbc_decrypt(ctx->__ctx_ptr,ctx->__decrypt,16,ctx->__xor_block,len,ciphertext,ciphertext);

	#endif

	return;
}


int hash_get_digest_size(int type) {

	switch (type) {
		case SHA1_160:
			return SHA1_160_DIGEST_SIZE;
		case TIGER_192:
			return TIGER_192_DIGEST_SIZE;
		case TIGER_160:
			return TIGER_160_DIGEST_SIZE;
		case TIGER_128:
			return TIGER_128_DIGEST_SIZE;
		default:
			return 0;
	}

}


int hash_get_block_size(int type) {
	/* both SHA1 and Tiger operate in 64 byte/512 bit blocks */
	return 64;
}


void hash_init(int type,hash_ctx *hash) {
	hash->type	= type;
	
	switch (type) {
		case SHA1_160:
			sha_init((SHA_CTX *)hash->ctx);
			hash->block_size	= SHA1_160_BLOCK_SIZE;
			hash->digest_size	= SHA1_160_DIGEST_SIZE;
			break;
		case TIGER_192:
			tiger_init((TIGER_CTX *)hash->ctx);
			hash->block_size	= TIGER_192_BLOCK_SIZE;
			hash->digest_size	= TIGER_192_DIGEST_SIZE;
			break;
		case TIGER_160:
			tiger_init((TIGER_CTX *)hash->ctx);
			hash->block_size	= TIGER_160_BLOCK_SIZE;
			hash->digest_size	= TIGER_160_DIGEST_SIZE;
			break;
		case TIGER_128:
			tiger_init((TIGER_CTX *)hash->ctx);
			hash->block_size	= TIGER_128_BLOCK_SIZE;
			hash->digest_size	= TIGER_128_DIGEST_SIZE;
			break;
		default:
			ERROR("bad hash type requested in hash_init");
	}	

	return; /* void */

} /* end hash_init */


void hash_update(hash_ctx *hash,void *plaintext, int size) {
	
	switch (hash->type) {
		case SHA1_160:
			sha_update((SHA_CTX *)hash->ctx,plaintext,size);
			break;
		case TIGER_192:
		case TIGER_160:
		case TIGER_128:
			tiger_update((TIGER_CTX *)hash->ctx,plaintext,size);
			break;
		default:
			ERROR("some huge error in hash_update!");
	}
	
	return; /* void */
	
} /* end hash_update */


void hash_final(hash_ctx *hash,void *digest) {

	switch (hash->type) {
		case SHA1_160:
			sha_final((SHA_CTX *)hash->ctx);
			sha_digest((SHA_CTX *)hash->ctx,digest);
			break;
		case TIGER_192:
			tiger_final((TIGER_CTX *)hash->ctx);
			tiger_digest((TIGER_CTX *)hash->ctx,digest);
			break;
		case TIGER_160:
			tiger_final((TIGER_CTX *)hash->ctx);
			tiger160_digest((TIGER_CTX *)hash->ctx,digest);
			break;
		case TIGER_128:
			tiger_final((TIGER_CTX *)hash->ctx);
			tiger128_digest((TIGER_CTX *)hash->ctx,digest);
			break;
		default:
			ERROR("some huge error in hash_final!");
	}
	
	return; /* void */

} /* hash_final */


void hmac_init(int type, hash_ctx *hash,void *key, int size) {
	int i;

	hash_init(type,hash);
	
	if (size > hash->block_size) {
		HASH(type,hash->key,key,size);
		hash->key_size = hash->digest_size;
	} else {
		/*bzero(hash->key,sizeof(hash->key));*/
		memcpy(hash->key,key,size);
		hash->key_size = size;
	}


	for (i = 0; i < hash->key_size; i++) {
		hash->ipad[i] = IPAD ^ hash->key[i];
	}

	for (; i < hash->block_size; i++) {
		hash->ipad[i] = IPAD;
	}


	hash_update(hash,hash->ipad,hash->block_size);

	return; /* void */

} /* end hmac_init */


#define hmac_update(hash,plaintext,size) \
	hash_update((hash),(plaintext),(size))


void hmac_final(hash_ctx *hash,void *digest) {
	int i;
	hash_ctx pad;
	
	for (i = 0; i < hash->key_size; i++) {
		hash->opad[i] = OPAD ^ hash->key[i];
	}

	for (; i < hash->block_size; i++) {
		hash->opad[i] = OPAD;
	}

	hash_init(hash->type,&pad);
	hash_update(&pad,hash->opad,hash->block_size);
	
	switch (hash->type) {
		case SHA1_160:
			sha_final((SHA_CTX *)hash->ctx);
			sha_digest((SHA_CTX *)hash->ctx,digest);
			break;
		case TIGER_192:
			tiger_final((TIGER_CTX *)hash->ctx);
			tiger_digest((TIGER_CTX *)hash->ctx,digest);
			break;
		case TIGER_160:
			tiger_final((TIGER_CTX *)hash->ctx);
			tiger160_digest((TIGER_CTX *)hash->ctx,digest);
			break;
		case TIGER_128:
			tiger_final((TIGER_CTX *)hash->ctx);
			tiger128_digest((TIGER_CTX *)hash->ctx,digest);
			break;
		default:
			ERROR("some huge error in hmac_final!");
	}

	hash_update(&pad,digest,hash->digest_size);
	hash_final(&pad,digest);

	return; /* void */

} /* end hmac_final */



/*
 * prand_init(*ctx,primitive,...)
 *
 * pseudo-randomness devices. this is used to build either
 * a PRG, pseudo-random generator (aka PRNG), or a PRF, psuedo-
 * random function.
 *
 * if the primitive is a cipher, then only a key and key_len are needed
 * to contruct the PRG.
 *
 * PRG example:
 *	prand_ctx ctx;
 * 	char key[]  = "my key";
 * 	prand_init(&ctx,ARCFOUR_STREAM,key,sizeof(key));
 *
 * if the primitive is a hash/one-way function, then both a key and
 * seed are needed to contruct the PRF.
 *
 * PRF example:
 * 	prand_ctx ctx;
 * 	char key[]  = "my key";
 * 	char seed[] = "my seed";
 * 	prand_init(&ctx,SHA1_160,key,sizeof(key),seed,sizeof(seed));
 *
 * --------------------------------------------------------------------------
 *
 * A PRF algorithm from TLS 1.0 RFC (absent the MD5/SHA1 split)
 * using an HMAC (keyed hashing) where HMAC like F(key,data).
 *
 * PRF(key, seed) = HMAC(key, A(0) + seed) +
 *                  HMAC(key, A(1) + seed) +
 *                  HMAC(key, A(2) + seed) + ...
 *                                                                 
 * Where + indicates concatenation.
 *                                                                    
 * A() is defined as:
 * A(0) = seed
 * A(i) = HMAC(key, A(i-1))
 *
 */
void prand_init(prand_ctx *ctx,int primitive,...) {
	va_list args;
	size_t key_len;
	size_t seed_len;
	size_t digest_size;
	unsigned char *key;
	unsigned char *seed;
	unsigned char *pool;

	va_start(args,primitive);		
	switch (primitive) {
		case ARCFOUR_STREAM:
			ctx->__primitive	= primitive;
			key			= va_arg(args,unsigned char *);
			key_len			= va_arg(args,size_t);
			arcfour_set_key(&(*ctx).__state.arcfour,key_len,key);
			break;
		case SHA1_160:
		case TIGER_192:
		case TIGER_160:
		case TIGER_128:
			digest_size = ctx->__state.hash.digest_size =
				hash_get_digest_size(primitive);
			ctx->__primitive	= primitive;
			key			= va_arg(args,unsigned char *);
			key_len			= va_arg(args,size_t);
			seed			= va_arg(args,unsigned char *);
			seed_len		= va_arg(args,size_t);

			ctx->__state.hash.key_len	= key_len;
			ctx->__state.hash.seed_len	= seed_len;
			ASSERT((ctx->__state.hash.key = _malloc(key_len)) != NULL,
				"malloc failed!");
			ASSERT((ctx->__state.hash.seed = _malloc(seed_len*2)) != NULL,
				"malloc failed!");

			memcpy(ctx->__state.hash.key,key,key_len);
			pool	= ctx->__state.hash.pool;

			/* initial: hmac(key,seed + seed); */
			memcpy(ctx->__state.hash.seed,seed,seed_len);
			memcpy(ctx->__state.hash.seed+seed_len,seed,seed_len);
			HMAC(primitive,pool,key,key_len,
				ctx->__state.hash.seed,seed_len*2);

			/* fill 2nd digest to normalize/fill the state */
			memcpy(ctx->__state.hash.seed,pool,digest_size);
			memcpy(ctx->__state.hash.seed+digest_size,seed,seed_len);
			HMAC(primitive,pool+digest_size,key,key_len,
				ctx->__state.hash.seed,digest_size+seed_len);

			ctx->__state.hash.pos		= 0;
			ctx->__state.hash.pool_size	= digest_size * 2;
			break;
		default:
			ABORT("bad primitive: %i!",primitive);
	}
	va_end(args);

	return;
} /* prand_init */


void prand_get(prand_ctx *ctx,void *buffer,size_t len) {
	size_t read,pool_size,digest_size,seed_len,key_len,*pos;
	unsigned char *seed,*key,*pool;
	unsigned char *buf = (unsigned char *)buffer;

	switch(ctx->__primitive) {
		case ARCFOUR_STREAM:
			arcfour_stream(&ctx->__state.arcfour,len,(uint8_t *)buf);
			break;
		case SHA1_160:
		case TIGER_192:
		case TIGER_160:
		case TIGER_128:
			seed		= ctx->__state.hash.seed;
			key		= ctx->__state.hash.key;
			pool		= ctx->__state.hash.pool;
			seed_len	= ctx->__state.hash.seed_len;
			key_len		= ctx->__state.hash.key_len;
			pool_size	= ctx->__state.hash.pool_size;
			digest_size	= ctx->__state.hash.digest_size;
			pos		= &ctx->__state.hash.pos;
			
			while (len > 0) {
				/* if its empty, fill w/ 2 digests;
				 * note: original seed never gets over written
				 * at &seed[digest_size] 
				 */
				if (*pos == pool_size) {

					/* fill 1st digest; 2nd digest is prev digest */
					memcpy(seed,pool+digest_size,digest_size);
					HMAC(ctx->__primitive,pool,key,key_len,seed,digest_size+seed_len);

					/* fill 2nd digest; 1st digest is prev digest */
					memcpy(seed,pool,digest_size);
					HMAC(ctx->__primitive,pool+digest_size,key,key_len,seed,digest_size+seed_len);

					/* replenished stream pool */
					*pos = 0;
				}

				read	= MIN((pool_size)-*pos,len);
				memcpy(buf,&pool[*pos],read);
				*pos 	+= read;
				buf 	+= read;
				len	-= read;
			}

			break;
		default:
			ABORT("bad primitive: %i!",ctx->__primitive);
	}

	return;
}


void prand_xor(prand_ctx *ctx,void *data,size_t len) {
	unsigned char buf[64];
	unsigned char *b = buf;
	unsigned char *d = (unsigned char *)data;
	ssize_t i, l = (ssize_t)len;

	while (l > 0) {
		prand_get(ctx,buf,MIN(l,(ssize_t)sizeof(buf)));
		MEMXOR(d,buf,MIN(l,(ssize_t)sizeof(buf)));
		d += sizeof(buf);
		l -= sizeof(buf);
	}
}


void prand_destroy(prand_ctx *ctx) {
	switch (ctx->__primitive) {
		case ARCFOUR_STREAM:
			break;
		case SHA1_160:
		case TIGER_192:
		case TIGER_160:
		case TIGER_128:
			_free(ctx->__state.hash.key);
			_free(ctx->__state.hash.seed);
			break;
		default:
			ABORT("bad primitive: %i!",ctx->__primitive);
	}

	return;
}


void dh_init(dh_ctx *dh) {

	mpz_init(dh->g); /* generator	*/
	mpz_init(dh->p); /* prime	*/
	mpz_init(dh->z); /* private key	*/
	mpz_init(dh->y); /* public key	*/
	mpz_init(dh->x); /* secret	*/

	return; /* void */

} /* end dh_init */


void dh_destroy(dh_ctx *dh) {

	mpz_clear(dh->g); /* generator		*/
	mpz_clear(dh->p); /* prime		*/
	mpz_clear(dh->x); /* private key	*/
	mpz_clear(dh->y); /* public key		*/
	mpz_clear(dh->z); /* secret		*/

	return; /* void */	

} /* end dh_destroy */


/*
 * generate size bits prime
 */
void dh_generate_parameters(dh_ctx *dh,int size,unsigned int generator) {

	//bignum_random_prime(dh->p,size);
	ke_get_prime(dh->p);
	mpz_set_ui(dh->g,generator);

	dh->p_size_bits		= mpz_sizeinbase(dh->p,2);
	dh->p_size_bytes	= bignum_format_u_length(dh->p);
	
	if (dh->p_size_bits != size)
		WARN("requested prime size does not seem to match existing: requested % i bits, have %i",size,dh->p_size_bits);
	
	return; /* void */

} /* end dh_generate_parameters */


void dh_set_prime(dh_ctx *dh,mpz_t p) {

	mpz_set(dh->p,p);
	
	dh->p_size_bytes	= bignum_format_u_length(dh->p);
	dh->p_size_bits		= mpz_sizeinbase(dh->p,2);

	return; /* void */

}


void dh_set_generator(dh_ctx *dh,unsigned int g) {

	//mpz_set(dh->g,g);
	mpz_set_ui(dh->g,g);
	
	return; /* void */

}


/*
 * generate public key
 *
 * y = g^x % p
 *
 */
void dh_generate_key(dh_ctx *dh) {
	unsigned char private[DH_MAX_PRIME_BYTE_SIZE];
	unsigned char less;
	int size;

	/* create secret */

	/*rand_get(&less,sizeof(less));
	size = (dh->p_size_bits-less)/8;
	rand_get(private,size);*/
	
	rand_get(private,dh->p_size_bytes - 1);
	bignum_parse_u(dh->x,dh->p_size_bytes - 1,private);

	mpz_powm(dh->y,dh->g,dh->x,dh->p);

	return; /* void */

} /* end dh_generate_key */


/*
 * generate shared secret
 *
 * z = y^x % p
 *
 */
void dh_compute_key(dh_ctx *dh,mpz_t y1) {

	mpz_powm(dh->z,y1,dh->x,dh->p);	

	return; /* void */

} /* end dh_compute_key */


/*
 * check dh session for bad parameters
 *
 * from [insert dhpaper.pdf name], section 7.1
 *
 * 1) g^x, g^y and g^xy != 1
 * 2) 1 < g^x, g^y < p - 1
 * 3) ... other checks
 *
 */
int dh_check(dh_ctx *dh) {
	return 0;
}


/*
 * key exchange initiation
 *
 */
void ke_init() {
	object_init(&__ke_mutex,NULL,NULL);
	schedule((void(*)(void*))ke_generate_prime,NULL,DEFAULT_PRIME_INTERVAL,false,true);
	mpz_init(__ke_prime);
	ke_load_prime();
	atexit(ke_save_prime);
}

void ke_generate_prime() {
	mpz_t p;
	
	mpz_init(p);
	bignum_random_prime(p,KE_PRIME_SIZE*8);

	object_lock(&__ke_mutex);

	mpz_set(__ke_prime,p);

	object_unlock(&__ke_mutex);

	INFO("generted new %i bit prime",KE_PRIME_SIZE*8);

	ke_save_prime();

	mpz_clear(p);

}

void ke_get_prime(mpz_t p) {
	object_lock(&__ke_mutex);
	mpz_set(p,__ke_prime);
	object_unlock(&__ke_mutex);
}

void ke_load_prime() {
	FILE *fp;
	bool generated = false;

	object_lock(&__ke_mutex);

	if ((fp = fopen(conf.prime,"r")) != NULL) {
		if (mpz_inp_raw(__ke_prime,fp) == 0) {
			ERROR("error in GMP loading stored prime from %s",conf.prime);
			INFO("generating new %i bit prime",KE_PRIME_SIZE*8);
			bignum_random_prime(__ke_prime,KE_PRIME_SIZE*8);
			generated = true;
		}
		INFO("loaded %i bit prime from %s",mpz_sizeinbase(__ke_prime,2),conf.prime);
		fclose(fp);
	} else {
		ERROR("error loading stored prime from %s: %s",conf.prime,strerror(errno));
		INFO("generating new %i bit prime",KE_PRIME_SIZE*8);
		bignum_random_prime(__ke_prime,KE_PRIME_SIZE*8);
		generated = true;
	}	

	object_unlock(&__ke_mutex);

	if (generated)
		ke_save_prime();
}

void ke_save_prime() {
	struct stat st;
	FILE *fp;

	object_lock(&__ke_mutex);

	INFO("saving prime to %s",conf.prime);
	if ((fp = fopen(conf.prime,"w")) != NULL) {
		if (mpz_out_raw(fp,__ke_prime) == 0) {
			ERROR("error in GMP saving stored prime to %s",conf.prime);
		}

		if (stat(conf.prime,&st) == 0) {
			if ((st.st_mode & S_IROTH) || (st.st_mode & S_IWOTH)) {
				WARN("promiscuous access to %s",conf.prime);
			}
		} else {
			ERROR("could not stat %s: %s",conf.prime,strerror(errno));
		}
		fclose(fp);
	} else {
		ERROR("error storing prime to %s: %s",conf.prime,strerror(errno));
	}	

	object_unlock(&__ke_mutex);

}


