/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <netinet/in.h>

#ifdef __USE_MCRYPT
#ifndef _MCRYPT_H
#define _MCRYPT_H
#include <mcrypt.h>
#endif
#endif

#include "common.h"
#include "thread.h"
#include "log.h"
#include "crypt.h"
#include "bignum.h"
#include "packet.h"
#include "node.h"


node_t *node_create(int sd) {
	node_t *node;

	if ((node = _malloc(sizeof(node_t))) == NULL) {
		ERROR("could not malloc space for a new node");
		return NULL;
	}
	bzero(node,sizeof(node_t));
	
	object_init(node,node_free,node);
	thread_init(node,node_run,node);

	node->east	= NULL;
	node->west	= NULL;
	node->channel	= NULL;

	if ((node->east = link_create(sd)) == NULL) {
		node_free(node);
		return NULL;
	}

	object_retain(node->east);

	__objects_created++;

	return node;
} /* end node_create */


void node_free(void *o) {
	node_t *n = (node_t *)o;

	thread_destroy(n);
	object_destroy(n);

	if (n->east != NULL)
		object_release(n->east);
	if (n->west != NULL)
		object_release(n->west);
	if (n->channel != NULL)
		object_release(n->channel);

	_free(n);

	__objects_freed++;
	
	return; /* void */
} /* end node_free */


void node_call_back(int sd) {
	node_t *node;
	
	if ((node = node_create(sd)) == NULL) {
		ERROR("unable to create node");
		return; /* void */
	}

	thread_run(node);
	
	return; /* void */

} /* node_call_back */


void node_run(void *o) {
	node_t *n	= (node_t *)o;	
	link_ctx *l	= NULL;
	channel_ctx *c	= NULL;

	DEBUG("node started");


	/*
	 * establish link
	 */
	link_respond(n->east);
	
	if (n->east->failed)
		goto fail;

	DEBUG("link established");


	/*
	 * establish channel
	 */
	if ((n->channel = channel_create(n->east)) == NULL)
		goto fail;

	object_retain(n->channel);
	channel_respond(n->channel);

	if (n->channel->failed)
		goto fail;
	
	DEBUG("channel established");

	node_loop(n);


	fail:

	node_close(n);

	return; /* void */
	
} /* node_run */


/*
 * NOTES:
 * 	fix link_poll to ignore LINK_FLUSHED on west link using a mask?
 */
void node_loop(node_t *n) {
	char st;
	link_ctx **east	= &n->east;
	link_ctx **west	= &n->west;
	channel_ctx **c	= &n->channel;
	packet_t *pkt;
	prand_ctx r;
	unsigned char e[16];
	int mark;

	rand_get(e,sizeof(e));
	prand_init(&r,ARCFOUR_STREAM,e,sizeof(e));


	for (;;) {
		st = link_poll(*east,*west);

		n->sent = false;
		mark	= (*c)->next_out;	/* keep track whether we */
						/* sent a packet or not */


		if (FLAG_ISSET(st,POLL_EFAILED) || FLAG_ISSET(st,POLL_WFAILED)) {
			link_fail(*east);
			link_fail(*west);
			break;
		} /*** else must be readable ***/


		/*
		 *  read channel
		 */
		if ((pkt = channel_get(*c)) == NULL) {
			if ((*c)->status == CHANNELFAILED) {
				link_fail(*east);
				link_fail(*west);
				break;
			}
		} else {
			/*** forward it? ***/
			if (FLAG_ISSET(PACKET_GET_FLAG(pkt->data,(*c)->hopid),PACKET_NODE_FWRD)) {
				/*** overwrite to end of our header ***/
				prand_get(&r,(pkt->data+1),
					(PACKET_HEAD(pkt->data,(*c)->hopid)+
					PACKET_HEAD_SIZE)-(pkt->data+1));

				if (*west != NULL) {
					link_put(*west,pkt);
				} else if ((*c)->reqflg != 0) {
					SIMPLEQ_INSERT_TAIL(&(*c)->reqfwd,pkt,sqe);
				} else {
					WARN("protocol violation: cannot forward packet to non-existant link");
					link_fail(*east);
					link_fail(*west);
					break;
				}

				node_cntl(n,NULL); /*** take care of asyc work ***/
			} else if (FLAG_ISSET(PACKET_GET_FLAG(pkt->data,(*c)->hopid),PACKET_NODE_NULL)) {
				link_cache_put(*east,pkt);
				node_cntl(n,NULL); /*** take care of async work ***/
			} else { /*** request/reply packet ***/
				node_cntl(n,pkt);
			}
		} /*** end read channel ***/


		/*
		 * read west link
		 */
		if (*west != NULL) {
			if ((pkt = link_get(*west)) != NULL) {
				channel_put(*c,pkt);
			} else if ((*c)->next_out == mark && FLAG_ISSET(st,POLL_EFLUSHED)) {
				/*** send NULL packet ***/
				if ((pkt = link_cache_get(*east)) != NULL) {
					PACKET_SET_FLAG(pkt->data,(*c)->hopid,PACKET_NODE_NULL);
					prand_get(&r,PACKET_DATA(pkt->data,(*c)->hopid),PACKET_DATA_SIZE);
					channel_put(*c,pkt);
				} else {
					link_fail(*east);
					link_fail(*west);
					break;
				}
			}
		} /*** read west link ***/


	} /* loop */


} /* node_loop */


void node_cntl(node_t *n,packet_t *pkt) {
	channel_ctx *c = n->channel;

	if (pkt != NULL) {
		/*
		 * is request buffer full, or req type different
		 * from current request operation
		 */
		if (BUFODD(&c->reqbuf) < PACKET_GET_PLEN(pkt->data,c->hopid)
			|| (c->reqflg != 0
			&& PACKET_GET_FLAG(pkt->data,c->hopid)
			!= c->reqflg)) {

			c->failed	= true;
			c->status	= CHANNELFAILED;
			return;
		} else { /*** just set; only useful for first packet ***/
			c->reqflg = PACKET_GET_FLAG(pkt->data,c->hopid);
		}

		c->reqpackets++;
		BUFCPY(&c->reqbuf,PACKET_DATA(pkt->data,c->hopid),
			PACKET_GET_PLEN(pkt->data,c->hopid));
		link_cache_put(c->link,pkt);
	}

	switch (c->reqflg) {
		case PACKET_NODE_CNCT:
			node_handle_connect(n);		
			break;
		default:
			c->failed	= true;
			c->status	= CHANNELFAILED;
	}

	return;

} /* node_cntl */


void node_close(node_t *n) {




} /* node_close */

#define RND MAX(1,((int)((n/m)+0.5)) + (1 * MIN(1,n%m)))
void node_handle_connect(node_t *n) {
	static const int max_request =
		1 + 255 + 1; /* flag + max domain (rfc2035) + nul */
	static const int max_packets =
		PACKET_BOUNDARY(1 + 255 + 1); /* flag + max domain + nul */

	//if (c->rebuf->data[0] == PACKET_CNCT_IPV4)

} /* node_handle_connect */


