/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <openssl/sha.h>
#include <openssl/evp.h>
#include <openssl/hmac.h>

#include "common.h"
#include "crypt.h"

/*
 * unsigned char *HMAC(const EVP_MD *evp_md,
 *                    const void *key, int key_len,
 *                    const unsigned char *d, int n,
 *                    unsigned char *md, unsigned int *md_len);
 */
 
/*
 * From TLS 1.0 Draft RFC (absent the MD5/SHA1 split)
 *
 * PRF(secret, seed) = HMAC_hash(secret, A(1) + seed) +
 *                     HMAC_hash(secret, A(2) + seed) +
 *                     HMAC_hash(secret, A(3) + seed) + ...
 *                                                                 
 * Where + indicates concatenation.
 *                                                                    
 * A() is defined as:
 * A(0) = seed
 * A(i) = HMAC_hash(secret, A(i-1))
 *
 */

/* PRF for key expansion */
void prf_expand(unsigned char *out, int out_len, const unsigned char *secret, int secret_len, const unsigned char *seed, int seed_len) {
	unsigned char md[EVP_MAX_MD_SIZE];
	unsigned char m[PRF_MAX_SEED_SIZE + MAX(PRF_MAX_SEED_SIZE,EVP_MAX_MD_SIZE)];
	unsigned char sc[PRF_MAX_SECRET_SIZE];
	unsigned char sd[PRF_MAX_SEED_SIZE];
	unsigned int md_len,m_len;
	int out_bytes = 0;
	unsigned char *prev;

	secret_len = MIN(PRF_MAX_SECRET_SIZE,secret_len);
	memcpy(sc,secret,secret_len);
	seed_len = MIN(PRF_MAX_SEED_SIZE,seed_len);
	memcpy(sd,seed,seed_len);

	/* round 1 */
	memcpy(&m[0],sd,seed_len);
	memcpy(&m[seed_len],sd,seed_len);
	m_len = seed_len*2;
	HMAC(EVP_sha1(),sc,secret_len,m,m_len,md,&md_len);

	if (out_len <= md_len) {
		memcpy(out,md,out_len);
		return;
	}

	memcpy(out,md,md_len);
	out += md_len;
	out_bytes += md_len;

	do {
		memcpy(m,md,md_len);
		memcpy((m+md_len),sd,seed_len);
		m_len = md_len + seed_len;
		HMAC(EVP_sha1(),sc,secret_len,m,m_len,md,&md_len);
		memcpy(out,md,MIN((out_len-out_bytes),md_len));
		out += md_len;
		out_bytes += md_len;
	} while (out_bytes < out_len);
	
}
